"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.evaluate = void 0;
const Exp = __importStar(require("../exp"));
const Stmt = __importStar(require("../stmt"));
const Value = __importStar(require("../value"));
const Neutral = __importStar(require("../neutral"));
const Closure = __importStar(require("../closure"));
const Telescope = __importStar(require("../telescope"));
const Mod = __importStar(require("../mod"));
const Env = __importStar(require("../env"));
const Trace = __importStar(require("../../trace"));
function evaluate(mod, env, exp, opts = {
    shadow_mod_value_p: false,
}) {
    try {
        switch (exp.kind) {
            case "Exp.v": {
                const value = Env.lookup(env, exp.name);
                if (value !== undefined)
                    return value;
                const mod_value = Mod.lookup_value(mod, exp.name);
                if (mod_value !== undefined) {
                    if (opts.shadow_mod_value_p) {
                        const entry = Mod.lookup_entry(mod, exp.name);
                        if (entry === undefined)
                            throw new Error("IMPOSSIBLE");
                        const t = Mod.lookup_type(mod, exp.name);
                        if (t === undefined)
                            throw new Error("IMPOSSIBLE");
                        if (entry.t !== undefined) {
                            Mod.update(mod, exp.name, {
                                ...entry,
                                value_cache: Value.not_yet(t, Neutral.v(exp.name)),
                            });
                        }
                    }
                    return mod_value;
                }
                throw new Trace.Trace(Exp.explain_name_undefined(exp.name));
            }
            case "Exp.pi": {
                return Value.pi(Exp.evaluate(mod, env, exp.arg_t, opts), Closure.create(mod, env, exp.name, exp.ret_t));
            }
            case "Exp.fn": {
                return Value.fn(Closure.create(mod, env, exp.name, exp.ret));
            }
            case "Exp.ap": {
                return Exp.do_ap(Exp.evaluate(mod, env, exp.target, opts), Exp.evaluate(mod, env, exp.arg, opts));
            }
            case "Exp.cls": {
                env = Env.clone(env);
                const sat = new Array();
                for (const entry of exp.sat) {
                    const name = entry.name;
                    const t = Exp.evaluate(mod, env, entry.t, opts);
                    const value = Exp.evaluate(mod, env, entry.exp, opts);
                    sat.push({ name, t, value });
                    Env.update(env, name, value);
                }
                if (exp.scope.length === 0) {
                    return Value.cls(sat, Telescope.create(mod, env, undefined, new Array()));
                }
                else {
                    const [entry, ...tail] = exp.scope;
                    const name = entry.name;
                    const t = Exp.evaluate(mod, env, entry.t, opts);
                    return Value.cls(sat, Telescope.create(mod, env, { name, t }, tail));
                }
            }
            case "Exp.fill": {
                const { target, arg } = exp;
                return Exp.do_fill(Exp.evaluate(mod, env, target, opts), Exp.evaluate(mod, env, arg, opts));
            }
            case "Exp.obj": {
                const { properties } = exp;
                return Value.obj(new Map(Array.from(properties, ([name, exp]) => [
                    name,
                    Exp.evaluate(mod, env, exp, opts),
                ])));
            }
            case "Exp.dot": {
                return Exp.do_dot(Exp.evaluate(mod, env, exp.target, opts), exp.name);
            }
            case "Exp.equal": {
                return Value.equal(Exp.evaluate(mod, env, exp.t, opts), Exp.evaluate(mod, env, exp.from, opts), Exp.evaluate(mod, env, exp.to, opts));
            }
            case "Exp.same": {
                return Value.same;
            }
            case "Exp.replace": {
                return Exp.do_replace(Exp.evaluate(mod, env, exp.target, opts), Exp.evaluate(mod, env, exp.motive, opts), Exp.evaluate(mod, env, exp.base, opts));
            }
            case "Exp.absurd": {
                return Value.absurd;
            }
            case "Exp.absurd_ind": {
                return Exp.do_absurd_ind(Exp.evaluate(mod, env, exp.target, opts), Exp.evaluate(mod, env, exp.motive, opts));
            }
            case "Exp.str": {
                return Value.str;
            }
            case "Exp.quote": {
                return Value.quote(exp.str);
            }
            case "Exp.union": {
                const { left, right } = exp;
                return Value.union(Exp.evaluate(mod, env, left, opts), Exp.evaluate(mod, env, right, opts));
            }
            case "Exp.type": {
                return Value.type;
            }
            case "Exp.begin": {
                env = Env.clone(env);
                for (const stmt of exp.stmts) {
                    Stmt.execute(mod, env, stmt);
                }
                return Exp.evaluate(mod, env, exp.ret, opts);
            }
            case "Exp.the": {
                return Exp.evaluate(mod, env, exp.exp, opts);
            }
        }
    }
    catch (error) {
        if (error instanceof Trace.Trace) {
            throw Trace.trail(error, exp);
        }
        else {
            throw error;
        }
    }
}
exports.evaluate = evaluate;
